# file: k8sAI/kuberag/tool_handler.py

import json
import os
from prompt_toolkit import prompt
from k8sAI.util import console
import subprocess


class ToolHandlerRegistry:
    """
    A registry for tool handlers.
    """

    def __init__(self):
        self.registry = {}

    def register_tool(self, prefix, handler):
        """
        Register a tool with a prefix and handler function.
        """
        if prefix in self.registry:
            raise ValueError(f"Prefix {prefix} is already registered.")
        self.registry[prefix] = handler

    def has_tool_handler(self, input_string):
        """
        Check if the input string has a registered handler.
        """
        for prefix in self.registry:
            if input_string.startswith(prefix):
                return True
        return False

    def use_handler(self, input_string):
        """
        Parse the input string and return the result.
        """
        for prefix, handler in self.registry.items():
            if input_string.startswith(prefix):
                return handler(input_string, prefix)
        console.print("Error: No handler found for the input string.")
        return None


# TODO: create proper tool handler objects with data parse fn, etc.


def handle_suggest_kubectl_tool(input_string, prefix) -> bool:
    """
    Handle the kubectl tool data.

    returns: True if the tool should terminate. False otherwise
    """
    cmd_and_output = input_string[len(prefix) :]
    cmd_and_output = (cmd_and_output.strip("'")).split("||")
    output, cmd = cmd_and_output[0], cmd_and_output[1]
    
    context = f"{cmd} was just run and it outputed {output}"

    return context, False


registry = ToolHandlerRegistry()
registry.register_tool("[Suggest_Kubectl_CMD_Tool]", handle_suggest_kubectl_tool)



# file: k8sAI/kuberag/retriever.py

from langchain_chroma import Chroma
from langchain_openai import OpenAIEmbeddings
from k8sAI.util import console


def load_retriever(embeddings_path: str):
    """
    Load a retriever from a given embeddings path.
    """
    vectordb = Chroma(
        persist_directory=embeddings_path,
        embedding_function=OpenAIEmbeddings(disallowed_special=()),
    )

    # Check if the vector database is empty and log a warning if so
    if len(vectordb.get(limit=1).get("documents", [])) == 0:
        console.print("Warning: The vector database is empty.")

    retriever = vectordb.as_retriever(
        search_type="mmr",  # Also test "similarity"
        search_kwargs={"k": 8},
    )
    return retriever



import configparser
from uuid import UUID
import uuid
import click
import os
import json
from posthog import Posthog
import requests


class Usage:
    def __init__(self):
        self.posthog = None
        self.enabled = False
        self.uuid = None

    def setup(self):
        self.prompt_usage()
        config_path = os.path.expanduser("~/.k8sAI/config.ini")
        config = configparser.ConfigParser()

        if os.path.exists(config_path):
            config.read(config_path)
            # Check if 'usage' section and necessary keys exist
            if "usage" in config and "usage_enabled" in config["usage"]:
                self.enabled = config["usage"].getboolean("usage_enabled")
            if "usage" in config and "uuid" in config["usage"]:
                self.uuid = config["usage"]["uuid"]
            if "key" in config:
                self.ph = config["key"]["ph"]

        if getattr(self, "enabled", False) and self.ph:
            self.posthog = Posthog(
                project_api_key=self.ph,
                host="https://us.i.posthog.com",
            )

    def prompt_usage(self):
        config = configparser.ConfigParser()
        config_path = os.path.expanduser("~/.k8sAI/config.ini")
        config.read(config_path)
        config_dir = os.path.dirname(config_path)
        if not os.path.exists(config_dir):
            os.makedirs(config_dir)

        if "usage" not in config:
            click.echo("Thank you for using k8sAI!")
            if click.confirm(
                "Do you want to help improve k8sAI by sending anonymous usage statistics?"
            ):
                # User agreed to usage
                config = configparser.ConfigParser()
                config.read(config_path)
                config["usage"] = {"usage_enabled": "true", "uuid": str(uuid.uuid4())}
                response = requests.get("http://k8s.wilsonspearman.com/getUsage")
                if response.status_code == 200:
                    ph = response.text
                    config["key"] = {"ph": ph}
                with open(config_path, "w") as configfile:
                    config.write(configfile)
                print(f"ph written to {config_path}")

                with open(config_path, "w") as configfile:
                    config.write(configfile)
                click.echo("Usage enabled. Thank you for your support!")
            else:
                # User declined usage
                config = configparser.ConfigParser()
                config["usage"] = {"usage_enabled": "false"}
                with open(config_path, "w") as configfile:
                    config.write(configfile)
                click.echo("Usage disabled. You can enable it later in the config.")

    def log_event(self, event_name):
        if self.posthog:
            self.posthog.capture(self.uuid, event_name)



# file: k8sAI/kuberag/chat.py

from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables.history import RunnableWithMessageHistory
from langchain.memory import ChatMessageHistory
from langchain_openai import ChatOpenAI
from langchain.agents import AgentExecutor, create_openai_functions_agent
from k8sAI.kuberag.tools import get_all_tools

llm = ChatOpenAI(model="gpt-4")

chat_prompt = ChatPromptTemplate.from_messages(
    [
        (
            "system",
            "You are a helpful, concise chatbot and kubernetes expert. You are also a fully functioning SRE with a toolset at your disposal. Use the tools you have to \
get information about kubernetes or to suggest a command to the user. \
When possible and if provided, use the execute tool to gather more information about the cluster before answering. \
If you use the execute tool, be sure to use the results in future answers and to fill in information and arguments. \
If the user is providing a task to you, consider using the execute tool to solve it. \
Refer to the chat history to check for relevant previous conversations and information. \
Users can't see your tool call results, so be sure to state all findings at the end clearly. \
If the user is requesting you to investigate a problem, use your knowledge as best you can to \
follow the best practices for troubleshooting k8s. When investigating a problem for a fix, make sure you utilize the your execute tool. \
If you find an issue, suggest a fix! If the user prompts you asking for a fix without a prompt, then take a wholistic look at the cluster health to try to find an issue. \
If nothing is wrong, and the user asks for a fix, concisely let them know everything looks healthy.\
Answer the user's questions based on your general knowledge of kubernetes, with your tools, \
and the following logs (if provided):\n\n{logs}\n\
Kubectl command output (optional):\n\n{command_output}",
        ),
        ("placeholder", "{chat_history}"),
        ("human", "{input}"),
        ("ai", "{agent_scratchpad}"),
    ]
)


def create_bot(retriever, disable_execution=False):
    """
    Create a bot with a retriever.
    """
    tools = get_all_tools(retriever, disable_execution=disable_execution)

    agent = create_openai_functions_agent(llm, tools, chat_prompt)
    agent_executor = AgentExecutor(agent=agent, tools=tools)

    memory = ChatMessageHistory(session_id="test-session")
    agent_with_chat_history = RunnableWithMessageHistory(
        agent_executor,
        lambda session_id: memory,
        input_messages_key="input",
        history_messages_key="chat_history",
    )

    return agent_with_chat_history



# file: k8sAI/main.py
"""
This module contains the commands for the CLI.
"""

from mimetypes import init
import subprocess
import click

from k8sAI.kuberag.main import k8sAI
from k8sAI.util import console, usage


@click.group()
def chat_group():
    """
    This tool uses OpenAI's ChatGPT to chat about Kubernetes.
    It is enhanced with a retriever storing k8s documentation to improve results.
    It uses your OpenAI API key from $OPENAI_API_KEY to access the model.

    NOTE: This tool sends data to OpenAI's servers. Please review the OpenAI API terms of use.
    """


@chat_group.command()
@click.option(
    "-p", "--prompt", help="Initial prompt to start the conversation. (optional)"
)
@click.option(
    "-t",
    "--terminal",
    is_flag=True,
    default=False,
    help="If terminal is enabled, k8sAI will end the conversation (optional)",
)
@click.option(
    "--disable-execution",
    is_flag=True,
    default=False,
    help="If execution is disabled, k8sAI wil not \
                be capable of executing kubectl commands (optional)",
)
def chat(prompt, terminal, disable_execution):
    """
    Start a conversation with the k8sAI model.
    k8sAI can suggest commands that can then be executed.
    """
    usage.log_event("chat_command")
    kube_ai = k8sAI(disable_execution)
    console.print("\n:robot: starting conversation with k8sAI...", style="bold green")
    if not terminal:
        console.print("   type 'exit' to end the conversation.", style="italic")

    kube_ai.start_chat(prompt, terminal=terminal)


@chat_group.command()
@click.option("--cmd", help="Kubectl command that k8sAI will explain the results of")
@click.option(
    "-p",
    "--prompt",
    help="Additional prompt to go along with command output,\
                otherwise default prompt is used (optional)",
)
@click.option(
    "-t",
    "--terminal",
    is_flag=True,
    default=False,
    help="If terminal is enabled, k8sAI will end the conversation (optional)",
)
@click.option(
    "--disable-execution",
    is_flag=True,
    default=False,
    help="If execution is disabled, k8sAI wil not be \
                capable of executing kubectl commands (optional)",
)
def explain(cmd, prompt, terminal, disable_execution):
    """
    kubectl command will be executed and k8sAI will explain the result
    """
    usage.log_event("explain_command")
    if not cmd.startswith("kubectl"):
        console.print("Error: Command must be kubectl command")
        return

    console.print(f":gear: running command: {cmd}")
    output = subprocess.check_output(cmd, shell=True).decode("utf-8")
    console.print(f":green_circle: command complete:\n{output}")

    if not prompt:
        prompt = "Concisely explain the output of the following command: " + cmd

    kube_ai = k8sAI(disable_execution)
    console.print("Explaining kubectl results with k8sAI...")
    if not terminal:
        console.print("Type 'exit' to end the conversation.")

    kube_ai.start_chat(prompt, command_output=output, terminal=terminal)


@chat_group.command()
@click.option("-p", "--prompt", help="A prompt describing the problem (optional)")
@click.option(
    "-t",
    "--terminal",
    is_flag=True,
    default=False,
    help="If terminal is enabled, k8sAI will end the conversation (optional)",
)
@click.option(
    "--disable-execution",
    is_flag=True,
    default=False,
    help="If execution is disabled, k8sAI wil not be capable of \
                executing kubectl commands (optional)",
)
def fix(prompt, terminal, disable_execution):
    """
    k8sAI will suggest a fix based on a description of the problem
    (or try to find a problem if none is provided)
    """
    usage.log_event("fix_command")
    enhanced_prompt = "Look for the root cause of the problem and suggest a fix."
    if prompt:
        enhanced_prompt += "The problem is:\n" + prompt
    else:
        enhanced_prompt += "Try to find the problem with your tools, \
            follow your best instincts for troubleshooting."

    kube_ai = k8sAI(disable_execution)
    console.print("Looking for fix with k8sAI...")
    if not terminal:
        console.print("Type 'exit' to end the conversation.")

    kube_ai.start_chat(enhanced_prompt, terminal=terminal)


def main():
    """
    Main function for the CLI.
    """
    usage.setup()
    chat_group()



# file: k8sAI/kuberag/tools.py

import json
import re
import subprocess
from typing import Optional, Type
from langchain.callbacks.manager import (
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)
from langchain.tools.retriever import create_retriever_tool
from langchain.pydantic_v1 import BaseModel, Field
from langchain.tools import BaseTool
from rich.padding import Padding
from k8sAI.util import console, usage
from prompt_toolkit import prompt
from prompt_toolkit.styles import Style


def retriever_tool(retriever):
    """
    Create the retriever tool.
    """
    return create_retriever_tool(
        retriever,
        "k8s_search",
        "Search for information about Kubernetes via the official documentation.\
            For any questions about Kubernetes, you must use this tool!",
    )


class SuggestKubectlCommandInput(BaseModel):
    """
    Suggest kubectl command input.
    """

    notes: str = Field(
        description="Any notes that the user should know about the command \
                       (a short docstring would be helpful)."
    )
    query: str = Field(
        description="should be a kubectl command that the user wants to run."
    )


class SuggestKubectlCommandTool(BaseTool):
    """
    Suggest kubectl command tool.
    """

    name = "Suggest_Kubectl_CMD_Tool"
    description = "Suggest a kubectl command to the user. \
        This is a terminal command and will allow the user to \
        execute the command directly. They will be able to edit it before executing.\
        The command MUST be properly formatted, but it can contain placeholder values for the user to fill in. \
        Fill in values to the BEST of your ability (but if you don't know, use a placeholder) and if you can retrieve them with another tool, do that first! \
        This is a useful tool for suggesting commands to the user and you're very much encouraged to use it \
        if it answer's the user's question or solves their errors. If the user's question is more general, you should use the `k8s_search` \
        tool instead or answer the question directly."
    args_schema: Type[BaseModel] = SuggestKubectlCommandInput
    return_direct: bool = True

    def _run(
        self,
        notes: str,
        query: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        usage.log_event("suggest_kubectl_command")
        if notes:
            console.print("\nNote: " + notes + "\n", style="italic")
        if query:
            cmd = prompt(
                "Edit the cmd and press (enter) to run, leave empty to return to GPT\n\n",
                default=query,
            )
        args = cmd.split()

        # Check if any command is entered, then execute it
        if args:
            console.print("\n")
            try:
                output = subprocess.check_output(cmd, shell=True).decode("utf-8")
                console.print(output)
            except Exception as e:
                output = str(e)
        else:
            console.print("No command entered.")

        return f"[Suggest_Kubectl_CMD_Tool]{output}||{cmd}"

    async def _arun(
        self,
        notes: str,
        query: str,
        run_manager: Optional[AsyncCallbackManagerForToolRun] = None,
    ) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError("Suggest Kubectl Command Tool does not support async")


class ExecuteKubectlCommandInput(BaseModel):
    command: str = Field(
        description="must be a properly formatted READ-ONLY kubectl command."
    )


class ExecuteKubectlCommandTool(BaseTool):
    name = "Execute_Kubectl_CMD_Tool"
    description = "Execute a kubectl command and receive the results. This allows you to retrieve info about the user's k8s \
    configuration in order to properly assit them. Anytime you need information about their cluster in order to inform \
    your decisions, use this tool by passing in a kubectl command. \
    When possible, use this tool over the suggest tool! Users appreciate that. \
    The results from this tool can be used to make better suggestions to the user, so this tool is extremely useful. \
    The command MUST be properly formatted. It can also only be a READ-ONLY command. \
    Under no circumstances should you modify underlying resources with this tool!!!"
    args_schema: Type[BaseModel] = ExecuteKubectlCommandInput
    return_direct: bool = False

    def _run(
        self, command: str, run_manager: Optional[CallbackManagerForToolRun] = None
    ) -> str:
        usage.log_event("execute_kubectl_command")
        # execute command
        if not command.startswith("kubectl"):
            return "Error: Command must be kubectl command"

        allowed_patterns = [
            r"^kubectl get ",
            r"^kubectl describe ",
            r"^kubectl explain ",
            r"^kubectl top ",
            r"^kubectl api-resources",
            r"^kubectl api-versions",
            r"^kubectl version",
            r"^kubectl config view",
            r"^kubectl logs ",
        ]
        if not any(re.match(pattern, command) for pattern in allowed_patterns):
            return "Error: Command is not permitted, must be a read-only operation"

        console.print(
            Padding(
                "k8sAI is requesting to run the following command:", pad=(0, 0, 0, 2)
            )
        )
        console.print(Padding(command + "\n", pad=(0, 0, 0, 4)), style="yellow")
        user_response = ""
        while user_response.lower() not in ["y", "n"]:
            user_response = input("  Enter y/n to approve/deny: ")
        if user_response.lower() == "n":
            usage.log_event("execute_kubectl_command_denied")
            return "User did not approve command execution. You can suggest the command instead."
        usage.log_event("execute_kubectl_command_approved")

        try:
            console.print(
                Padding(f":gear: running command: {command}", pad=(0, 0, 0, 2))
            )
            output = subprocess.check_output(command, shell=True).decode("utf-8")
            console.print(Padding(":green_circle: command complete", pad=(0, 0, 0, 2)))
            return output
        except subprocess.CalledProcessError as e:
            return f"Error: {e}"
        except Exception as e:
            return f"Error: {e}"

    async def _arun(
        self,
        command: str,
        run_manager: Optional[AsyncCallbackManagerForToolRun] = None,
    ) -> str:
        """Use the tool asynchronously."""
        raise NotImplementedError("Suggest Kubectl Command Tool does not support async")


def get_all_tools(retriever, disable_execution=False):
    """
    Get all tools.
    """
    tools = [retriever_tool(retriever), SuggestKubectlCommandTool()]
    if not disable_execution:
        tools.append(ExecuteKubectlCommandTool())
    return tools



# file: k8sAI/kuberag/main.py
import os
import pkg_resources
from rich.padding import Padding
from rich.markdown import Markdown

from k8sAI.kuberag.chat import create_bot
from k8sAI.kuberag.retriever import load_retriever
from k8sAI.kuberag.tool_handler import registry
from k8sAI.util import console


def get_embeddings_path():
    """Get the absolute path to the embeddings directory."""
    resource_path = pkg_resources.resource_filename(__name__, "embeddings")
    return resource_path


class k8sAI:
    """
    k8sAI is a class that allows you to chat with a GPT model.
    It uses a retriever to find relevant information.
        Args:
            disable_execution (bool): If True, the bot cannot execute kubectl commands to help.
            embeddings_path (str): The path to the stored embeddings
    """

    def __init__(
        self,
        disable_execution: bool = False,
        embeddings_path: str = get_embeddings_path(),
    ):
        self.retriever = load_retriever(embeddings_path)
        self.bot = create_bot(self.retriever, disable_execution)

    def start_chat(
        self,
        prompt,
        command_output="",
        logs: str = "<no logs provided>",
        terminal=False,
    ):
        """
        Start a chat with the bot. User can end with 'exit'.
            Args:
                logs (str): The logs to provide to the bot.
        """
        additional_context = None
        while True:
            if prompt:
                user_prompt = prompt
                prompt = None
            else:
                user_prompt = input("Prompt: ")
            if user_prompt.lower() == "exit":
                break
            # result = self.bot.invoke()
            if additional_context != None:
                user_prompt = f"using this \n{additional_context} \nrespond to this: {user_prompt}"
                additional_context = None
            for chunk in self.bot.stream(
                {"input": user_prompt, "logs": logs, "command_output": command_output},
                config={"configurable": {"session_id": "<foo>"}},
            ):
                # Agent Action
                if "actions" in chunk:
                    for action in chunk["actions"]:
                        console.print(
                            Padding(
                                f":hammer: Calling Tool: `{action.tool}` with input `{action.tool_input}`",
                                pad=(0, 0, 0, 2),
                            ),
                            style="italic bright_black",
                        )
                # Observation
                elif "steps" in chunk:
                    continue
                # Final result
                elif "output" in chunk:
                    if registry.has_tool_handler(chunk["output"]):
                        additional_context, terminate = registry.use_handler(
                            chunk["output"]
                        )
                        if terminate:
                            break
                    else:
                        console.print("k8sAI:\n", style="bold green")
                        console.print(Markdown(chunk["output"]))

                else:
                    raise ValueError()
                console.print("---")

            if terminal:
                break


if __name__ == "__main__":
    path = os.path.join(os.getcwd(), "k8sAI/kuberag/embeddings")
    k8sAI = k8sAI(embeddings_path=path)
    k8sAI.start_chat(None)



from setuptools import setup, find_packages
from setuptools.command.install import install

setup(
    name="k8sAI",
    version="0.1.7",
    packages=find_packages(),
    description="A conversational AI for Kubernetes users.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Wilson Spearman",
    author_email="wilsonspearman@gmail.com",
    url="https://github.com/wilson090/k8sAI",
    include_package_data=True,
    install_requires=[
        "click",
        "langchain",
        "langchain-chroma",
        "langchain_openai",
        "prompt_toolkit",
        "rich",
        "posthog",
    ],
    entry_points="""
        [console_scripts]
        k8sAI=k8sAI.main:main
    """,
)
